from django.test import TestCase as DjangoTestCase


class TestCase(DjangoTestCase):
    def __init__(self, *args, **kwargs):
        if not hasattr(self, 'assertItemsEqual'):
            self.assertItemsEqual = self.assertCountEqual
        super(TestCase, self).__init__(*args, **kwargs)



from collections import OrderedDict

from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.views.generic.base import TemplateView

from . import app_settings

try:
    from django.urls import re_path
except ImportError:
    from django.conf.urls import url as re_path



class ControlCenter(object):
    def __init__(self, name, view_class):
        self.name = name
        self.view_class = view_class

    def get_view(self):
        return self.view_class.as_view(controlcenter=self)

    def get_urls(self):
        urlpatterns = [
            re_path(r'^$', self.get_view(), name='index'),
            re_path(r'^(?P<pk>\w+)/$', self.get_view(), name='dashboard'),
        ]
        return urlpatterns

    @property
    def urls(self):
        return self.get_urls(), 'controlcenter', self.name


class DashboardView(TemplateView):
    dashboard = None
    controlcenter = None
    template_name = 'controlcenter/dashboard.html'

    @method_decorator(staff_member_required)
    def dispatch(self, *args, **kwargs):
        return super(DashboardView, self).dispatch(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        pk = self.kwargs.get('pk')

        # Redirects to the first dashboard if pk is not provided
        if not pk and self.dashboards:
            dashboard = next(iter(self.dashboards.values()))
            return redirect(dashboard.get_absolute_url())

        try:
            self.dashboard = self.dashboards[pk]
        except KeyError:
            raise Http404(f'Dashboard "{pk}" not found')
        return super(DashboardView, self).get(request, *args, **kwargs)

    @cached_property
    def dashboards(self):
        dashboards = OrderedDict()
        for slug, path in enumerate(app_settings.DASHBOARDS):
            if isinstance(path, (list, tuple)):
                slug, path = path
            pk = str(slug)
            klass = import_string(path)
            dashboards[pk] = klass(pk=pk)

        if not dashboards:
            raise ImproperlyConfigured('No dashboards found.')
        return dashboards

    def get_context_data(self, **kwargs):
        context = {
            'title': self.dashboard.title,
            'dashboard': self.dashboard,
            'dashboards': self.dashboards.values(),
            'groups': self.dashboard.get_widgets(self.request),
            'sharp': app_settings.SHARP,
        }

        # Admin context
        kwargs.update(admin.site.each_context(self.request))
        kwargs.update(context)
        return super(DashboardView, self).get_context_data(**kwargs)


controlcenter = ControlCenter('controlcenter', DashboardView)



from controlcenter.widgets.contrib import simple

from . import TestCase


class ValueListTest(TestCase):
    def setUp(self):
        self.widget = ExampleValueList(request=None)

    def test_basic(self):
        self.assertIsNotNone(self.widget.template_name)

    def test_default_not_sortable(self):
        self.assertFalse(self.widget.sortable)

    def test_get_data(self):
        self.assertItemsEqual(self.widget.items(), ['Label 1', 'Label 2'])


class KeyValueListTest(TestCase):
    def setUp(self):
        self.widget = ExampleKeyValueList(request=None)

    def test_basic(self):
        self.assertIsNotNone(self.widget.template_name)

    def test_default_not_sortable(self):
        self.assertFalse(self.widget.sortable)

    def test_get_data(self):
        self.assertItemsEqual(
            self.widget.items(),
            {'Key 1': 'Value 1', 'Key 2': 'Value 2'}.items(),
        )


class ExampleValueList(simple.ValueList):
    title = 'Value list widget'

    def get_data(self):
        return ['Label 1', 'Label 2']


class ExampleKeyValueList(simple.KeyValueList):
    title = 'Key-value list widget'

    def get_data(self):
        return {'Key 1': 'Value 1', 'Key 2': 'Value 2'}



import json
from collections.abc import Sequence, Mapping
from functools import partial

from django import template
from django.core.exceptions import FieldDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models.base import ModelBase
from django.urls import NoReverseMatch, reverse
from django.utils.html import conditional_escape, format_html, mark_safe
from django.utils.http import urlencode

from .. import app_settings
from ..utils import indexonly

register = template.Library()


@register.filter
def jsonify(obj):
    return mark_safe(json.dumps(obj, cls=DjangoJSONEncoder))


@register.filter
def is_sequence(obj):
    return isinstance(obj, Sequence)


@register.simple_tag
def change_url(widget, obj):

    if not widget.model and not isinstance(obj, models.Model):
        # No chance to get model url
        return

    elif isinstance(obj, Mapping):
        pk = obj.get('pk', obj.get('id'))
        meta = widget.model._meta

    elif indexonly(obj):
        if not widget.list_display:
            # No chance to guess pk
            return
        # Excludes sharp tip zip keys and values
        keys = (k for k in widget.list_display if k != app_settings.SHARP)
        new_obj = {x: y for x, y in zip(keys, obj)}
        return change_url(widget, new_obj)
    else:
        pk = getattr(obj, 'pk', getattr(obj, 'id', None))
        if not isinstance(obj, models.Model):
            # Namedtuples and custom stuff
            meta = widget.model._meta
        elif getattr(obj, '_deferred', False):  # pragma: no cover
            # Deferred model
            meta = obj._meta.proxy_for_model._meta
        else:
            # Regular model or django 1.10 deferred
            meta = obj._meta

    if pk is None:
        # No chance to get item url
        return

    try:
        app_label, model_name = meta.app_label, meta.model_name
        return reverse('admin:{}_{}_change'.format(app_label, model_name),
                       args=[pk])
    except NoReverseMatch:
        return


@register.filter
def changelist_url(widget):
    obj = widget.changelist_url
    if not obj or isinstance(obj, str):
        # If it's None (empty) or provided a real url
        return obj
    elif isinstance(obj, (tuple, list)):
        # Model, {'key1': 'value1', 'key2':...}
        # Model, '?key1=value1&key2=value2...'
        model, params = obj
    else:
        # Just a Model
        model, params = obj, ''

    assert isinstance(model, ModelBase), (
        '{}.changelist_url should be a string or Model class, or a tuple '
        'with Model class as the first element.'.format(widget))
    assert isinstance(params, (dict, str)), (
        'Widget.changelist_url\'s second element should be a dict or '
        'a string.')

    if isinstance(params, dict):
        urlparams = urlencode(params, doseq=True)
    else:
        # It's a string
        urlparams = params

    if urlparams and not urlparams.startswith('?'):
        urlparams = '?' + urlparams

    changelist_url = reverse('admin:{}_{}_changelist'
                             .format(model._meta.app_label,
                                     model._meta.model_name))
    return changelist_url + urlparams


@register.simple_tag
def attrvalue(widget, obj, attrname):
    """
    Looks for an attribute value in widget.
    Then tries to get one from object (dict, sequence, model,
    namedtuple, whatever looks alike).
    """

    # Looks for a callable in widget
    attr = getattr(widget, attrname, None)
    if attr and callable(attr):
        value = attr(obj)
    elif isinstance(obj, Mapping):
        # Manager.values()
        value = obj.get(attrname)
    elif indexonly(obj):
        # Fist removes sharp from the list
        clone = list(widget.list_display or ())
        if app_settings.SHARP in clone:
            clone.remove(app_settings.SHARP)

        if attrname not in clone:
            value = None
        else:
            index = clone.index(attrname)
            try:
                # Obj might be shorter
                value = obj[index]
            except IndexError:
                value = None
    else:
        # Model, namedtuple, custom stuff
        attr = getattr(obj, attrname, None)
        value = attr() if callable(attr) else attr

    if value is None:
        # It's not found or the value is None
        return ''
    elif attr and getattr(attr, 'allow_tags', False):
        return mark_safe(value)
    else:
        # TODO: django 1.9 auto-escapes simple_tag
        return conditional_escape(value)


def _method_prop(obj, attrname, attrprop):
    """
    Returns property of callable object's attribute.
    """
    attr = getattr(obj, attrname, None)
    if attr and callable(attr):
        return getattr(attr, attrprop, None)


_method_label = partial(_method_prop, attrprop='short_description')


@register.filter
def attrlabel(widget, attrname):
    widget_prop = _method_label(widget, attrname)
    if widget_prop is not None:
        return widget_prop

    elif widget.model:
        model_prop = _method_label(widget.model, attrname)
        if model_prop is not None:
            # Allows to have empty description
            return model_prop

        if attrname == 'pk':
            fieldname = widget.model._meta.pk.name
        else:
            fieldname = attrname

        try:
            field = widget.model._meta.get_field(fieldname)
            return field.verbose_name
        except FieldDoesNotExist:
            pass
    return attrname


@register.simple_tag
def external_link(url, label=None):
    return format_html(
        '<a href="{href}" target="_blank" '
        'rel="noreferrer" rel="noopener">{label}</a>',
        href=url, label=label or url,
    )



from .utils import captitle


class BaseModel(object):
    title = None

    def __init__(self):
        self.__name__ = self.__class__.__name__
        self.slug = self.__name__.lower()
        if self.title is None:
            self.title = captitle(self.__name__)

    def __str__(self):
        return self.__name__



from django.db import models


class TestUser0(models.Model):
    username = models.CharField(max_length=255)
    test_field = models.CharField('My title', max_length=255)

    class Meta:
        app_label = 'controlcenter'

    def foo(self):
        return 'original foo value'
    foo.short_description = 'original foo label'

    def bar(self):
        return 'original bar value'
    bar.short_description = 'original bar label'

    def baz(self):
        pass
    baz.short_description = ''

    def egg(self):
        return 'original egg value'


class TestUser1(models.Model):
    primary = models.AutoField(primary_key=True)
    username = models.CharField(max_length=255)

    class Meta:
        app_label = 'controlcenter'



import collections
from collections.abc import Sequence

from django.contrib.auth.models import User

from controlcenter.utils import captitle, deepmerge, indexonly

from . import TestCase


class SampleTest(TestCase):
    def test_captitle(self):
        title = captitle('TestClass')
        self.assertEqual(title, 'Test class')

    def test_deepmerge(self):
        first = {
            'a': 0,
            'b': {
                'c': {
                    'd': 0,
                    'e': 0
                }
            },
            'f': 0
        }

        second = {
            'a': 1,
            'b': {
                'c': {
                    'd': 1
                }
            },
            'g': 1
        }

        # Just to make sure
        with self.assertRaises(AssertionError):
            self.assertDictEqual(first, second)

        # Merge
        merged = deepmerge(first, second)

        # Checks merged
        self.assertItemsEqual(merged, ['a', 'b', 'f', 'g'])
        self.assertEqual(merged['a'], 1)
        self.assertEqual(merged['b']['c']['d'], 1)
        self.assertEqual(merged['b']['c']['e'], 0)
        self.assertEqual(merged['f'], 0)
        self.assertEqual(merged['g'], 1)

        # First wasn't changed
        self.assertItemsEqual(first, ['a', 'b', 'f'])
        self.assertEqual(first['b']['c']['d'], 0)
        with self.assertRaises(KeyError):
            first['g']

        # Second wasn't changed
        self.assertItemsEqual(second, ['a', 'b', 'g'])
        self.assertEqual(second['b']['c']['d'], 1)
        with self.assertRaises(KeyError):
            second['f']

        # Inner dict check
        with self.assertRaises(KeyError):
            second['b']['c']['e']

    def test_indexonly(self):
        # Built-in types
        self.assertTrue(indexonly(list()))
        self.assertTrue(indexonly(tuple()))
        self.assertFalse(indexonly(dict()))

        # Django model
        self.assertFalse(indexonly(User()))

        class MySequence(Sequence):
            def __init__(self, data):
                self.data = data

            def __getitem__(self, index):
                return self.data[index]

            def __len__(self):
                return len(self.data)

        MyTuple = collections.namedtuple('MyTuple', 'x y')

        # Custom stuff
        self.assertTrue(indexonly(MySequence([0, 1])))
        self.assertFalse(indexonly(MyTuple(0, 1)))



SECRET_KEY = 'test'
DEBUG = True
ALLOWED_HOSTS = []
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'controlcenter',
)
MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

# https://docs.djangoproject.com/en/4.0
# /releases/3.2/#customizing-type-of-auto-created-primary-keys
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
ROOT_URLCONF = 'urls'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        # 'NAME': os.path.join(BASE_DIR, 'test.db'),
    }
}
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'



from collections.abc import Sequence

from django.utils.text import camel_case_to_spaces, capfirst

__all__ = ['captitle', 'deepmerge']


def captitle(title):
    return capfirst(camel_case_to_spaces(title))


def deepmerge(*dicts):
    merged = {}
    for dct in dicts:
        for key, value in dct.items():
            if not isinstance(value, dict):
                merged[key] = value
            else:
                origin = merged.get(key, {})
                merged[key] = deepmerge(origin, value)
    return merged


def indexonly(obj):
    # Checks if given object is a sequence but not a namedtuple
    # If your custom class instance passed this, then why did you do that?
    return (isinstance(obj, Sequence) and not
            hasattr(obj, '_make') and not hasattr(obj, '_replace'))



from collections.abc import Sequence
import itertools
import os
from abc import ABCMeta

from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property

from ..base import BaseModel

__all__ = ['Group', 'ItemList', 'Widget', 'SMALL', 'MEDIUM', 'LARGE',
           'LARGER', 'LARGEST', 'FULL']


# Actually we don't need all that sizes
# but should have a grid for Masonry
# so I'm going to leave the most helpful ones

SMALL = 1    # 25%  or  [x] + [x] + [x] + [x]
MEDIUM = 2   # 33%  or  [ x ] + [ x ] + [ x ]
LARGE = 3    # 50%  or  [   x   ] + [   x   ]
LARGER = 4   # 66%  or  [     x     ] + [ x ]
LARGEST = 5  # 75%  or  [      x      ] + [x]
FULL = 6     # 100% or  [         x         ]


class WidgetMeta(ABCMeta):
    # Makes certain methods cached
    CACHED_ATTRS = (
        # values for charts and itemlists
        'values',
    )

    def __new__(mcs, name, bases, attrs):
        # We can't use it on __init__ because
        # cached_property fires on property's __get__
        for attr in mcs.CACHED_ATTRS:
            if attr in attrs:
                attrs[attr] = cached_property(attrs[attr])
        return super(WidgetMeta, mcs).__new__(mcs, name, bases, attrs)


class BaseWidget(BaseModel, metaclass=WidgetMeta):
    title = None
    model = None
    queryset = None
    changelist_url = None
    cache_timeout = None
    template_name = None
    template_name_prefix = None
    limit_to = None
    width = None
    height = None

    def __init__(self, request, **options):
        super(BaseWidget, self).__init__()
        self.request = request
        self.init_options = options

    def get_template_name(self):
        assert self.template_name, (
            '{}.template_name is not defined.'.format(self))
        return os.path.join(self.template_name_prefix.rstrip(os.sep),
                            self.template_name.lstrip(os.sep))

    def get_queryset(self):
        # Copied from django.views.generic.detail
        # Boolean check will run queryset
        if self.queryset is not None:
            return self.queryset.all()
        elif self.model:
            return self.model._default_manager.all()
        raise ImproperlyConfigured(
            '{name} is missing a QuerySet. Define '
            '{name}.model, {name}.queryset or override '
            '{name}.get_queryset().'.format(name=self.__class__.__name__))

    def values(self):
        # If you put limit_to in get_queryset method
        # using of super().get_queryset() will not make any sense
        # because the queryset will be sliced
        queryset = self.get_queryset()
        if self.limit_to:
            return queryset[:self.limit_to]
        return queryset


class Group(Sequence):
    def __init__(self, widgets=None, attrs=None, width=None, height=None):
        self.widgets = tuple(widgets or ())
        self.attrs = (attrs or {}).copy()
        self.width, self.height = width, height

    def __repr__(self):
        return '<Group of widgets: {}>'.format(self.widgets)

    def __len__(self):
        return len(self.widgets)

    def __getitem__(self, index):
        return self.widgets[index]

    def __add__(self, other):
        widgets = itertools.chain(self, other)
        width = getattr(other, 'width', self.width)
        height = getattr(other, 'height', self.height)
        attrs = self.attrs.copy()
        other_attrs = getattr(other, 'attrs', None)
        if other_attrs:
            attrs.update(other_attrs)
        return Group(widgets, attrs, width, height)

    def get_id(self):
        return self.attrs.get('id', '_and_'.join(x.slug for x in self))

    def get_class(self):
        return self.attrs.get('class', '')

    def get_attrs(self):
        return {key: value for key, value
                in self.attrs.items()
                if key not in ('id', 'class')}

    def _get_size(self, size):
        value = getattr(self, size, None)
        if value is not None:
            return value
        elif self.widgets:
            # py3 hooks
            return max(getattr(x, size, None) or 0 for x in self) or None

    def get_width(self):
        return self._get_size('width')

    def get_height(self):
        return self._get_size('height')


class Widget(BaseWidget):
    limit_to = 10  # It's always a good reason to limit queryset
    width = MEDIUM
    template_name_prefix = 'controlcenter/widgets'


class ItemList(Widget):
    list_display = None
    list_display_links = None
    template_name = 'itemlist.html'
    empty_message = 'No items to display'
    sortable = False



import itertools

from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.db.models.query import QuerySet

from controlcenter import widgets
from controlcenter.widgets.core import BaseWidget, WidgetMeta

from . import TestCase


class WidgetMetaTest(TestCase):
    def test_property_cache(self):
        class TestWidget(metaclass=WidgetMeta):
            def values(self):
                return next(itertools.count())

        widget = TestWidget()

        # self.values is a cached property
        for i in range(2):
            self.assertEqual(widget.values, 0)


class BaseWidgetTest(TestCase):
    def setUp(self):
        self.options = {'foo': 'bar'}
        self.widget0 = BaseWidget(request=None, **self.options)
        self.widget1 = BaseWidget(request=None, **self.options)

        for i in range(10):
            username = 'user{}'.format(i)
            User.objects.create_user(username, username + '@example.com',
                                     username + 'password')

    def test_init(self):
        # First argument is request
        with self.assertRaises(TypeError):
            BaseWidget()

        # Init options
        self.assertEqual(self.widget0.init_options, self.options)

    def test_template_name(self):
        # No template_name
        with self.assertRaises(AssertionError):
            self.widget0.get_template_name()

        # Simple test
        self.widget0.template_name_prefix = 'prefix'
        self.widget0.template_name = 'test.html'
        self.assertEqual(self.widget0.get_template_name(),
                         'prefix/test.html')

        # Slash test
        self.widget0.template_name_prefix = 'prefix/////'
        self.widget0.template_name = '//test.html'
        self.assertEqual(self.widget0.get_template_name(),
                         'prefix/test.html')

    def test_queryset(self):
        # No queryset was provided
        with self.assertRaises(ImproperlyConfigured):
            self.widget0.get_queryset()

        # Model is defined, queryset is NOT
        self.widget0.model = User
        self.assertItemsEqual(self.widget0.get_queryset(),
                              User._default_manager.all())

        # Both are defined
        self.widget0.queryset = User.objects.filter(username='user0')
        self.assertItemsEqual(self.widget0.get_queryset(),
                              User._default_manager.filter(username='user0'))

    def test_values(self):
        # Setup
        self.widget0.model = User
        self.widget0.limit_to = limit = 4

        # It's not a bound method or something
        self.assertIsInstance(self.widget0.values, QuerySet)

        # Limits values
        self.assertEqual(len(self.widget0.values), limit)

        # Drops limit
        self.widget1.model = User
        self.widget1.limit_to = None
        self.assertEqual(len(self.widget1.values), User.objects.count())


class ItemListTest(TestCase):
    # TODO: template test

    def setUp(self):
        self.widget = widgets.ItemList(request=None)

    def test_basic(self):
        # Required attributes
        self.assertIsNotNone(self.widget.template_name)
        self.assertIsNotNone(self.widget.empty_message)


class GroupTest(TestCase):
    def setUp(self):
        self.widget0 = widgets.ItemList(request=None)
        self.widget1 = widgets.ItemList(request=None)
        self.widget1.width = widgets.LARGE
        self.widget1.height = 300

    def test_empty_group(self):
        group = widgets.Group()

        # It's empty and immutable
        self.assertEqual(group.widgets, ())

        # __iter__ test
        self.assertItemsEqual(group, [])
        self.assertEqual(len(group), 0)

        # Attributes test
        self.assertEqual(group.get_id(), '')
        self.assertEqual(group.get_class(), '')
        self.assertEqual(group.get_attrs(), {})
        self.assertIsNone(group.get_width())
        self.assertIsNone(group.get_height())

    def test_nonempty_group(self):
        group = widgets.Group([self.widget0])

        # It's empty and immutable
        self.assertEqual(group.widgets, (self.widget0,))

        # __iter__
        self.assertItemsEqual(group, (self.widget0,))
        self.assertEqual(len(group), 1)

        # Attributes test
        self.assertEqual(group.get_id(), 'itemlist')
        self.assertEqual(group.get_class(), '')
        self.assertEqual(group.get_attrs(), {})
        self.assertEqual(group.get_width(), widgets.Widget.width)
        self.assertIsNone(group.get_height())

    def test_concatenation(self):
        group = (widgets.Group([self.widget0]) +
                 widgets.Group([self.widget1],
                               attrs={'class': 'clearfix', 'data-prop': '2'}))

        # It's empty and immutable
        self.assertEqual(group.widgets, (self.widget0, self.widget1))

        # __iter__
        self.assertItemsEqual(group, (self.widget0, self.widget1))
        self.assertEqual(len(group), 2)

        # Attributes test
        self.assertEqual(group.get_id(), 'itemlist_and_itemlist')
        self.assertEqual(group.get_class(), 'clearfix')
        self.assertEqual(group.get_attrs(), {'data-prop': '2'})
        self.assertEqual(group.get_width(), widgets.LARGE)
        self.assertEqual(group.get_height(), 300)

        # Unknown attr
        self.assertIsNone(group._get_size('length'))

        # Overwrite width
        self.assertEqual(widgets.Group(width=500).get_width(), 500)

        # test repr
        self.assertEqual(repr(group),
                         '<Group of widgets: {}>'
                         .format((self.widget0, self.widget1)))



import collections
import json

from django import VERSION
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType

from controlcenter import app_settings, widgets
from controlcenter.templatetags.controlcenter_tags import (
    _method_prop,
    attrlabel,
    attrvalue,
    change_url,
    changelist_url,
    external_link,
    is_sequence,
    jsonify,
)
from test_models import TestUser0, TestUser1

from . import TestCase


class SimpleTagsTest(TestCase):
    def test_jsonify(self):
        data = {'a': None, 'b': 0}
        json_data = jsonify(data)

        # Marked safe
        self.assertTrue(hasattr(json_data, '__html__'))
        self.assertEqual(json_data, json.dumps(data))

    def test_is_sequence(self):
        self.assertTrue(is_sequence(list()))
        self.assertTrue(is_sequence(tuple()))
        self.assertFalse(is_sequence(dict()))
        self.assertFalse(is_sequence(User()))

    def test_changelist_url(self):
        widget = widgets.ItemList(request=None)
        widget.changelist_url = 'test'

        # Original
        admin_changelist_url = '/admin/auth/user/'

        # String test
        self.assertEqual(changelist_url(widget), 'test')

        # Model test
        widget.changelist_url = User
        self.assertEqual(changelist_url(widget),
                         admin_changelist_url + '')

        # Tuple with params test
        widget.changelist_url = (User, {'username__exact': 'user0'})
        self.assertEqual(changelist_url(widget),
                         admin_changelist_url + '?username__exact=user0')

        # Same with string no question sign
        widget.changelist_url = (User, 'username__exact=user0')
        self.assertEqual(changelist_url(widget),
                         admin_changelist_url + '?username__exact=user0')

        # Same with question sign
        widget.changelist_url = (User, '?username__exact=user0')
        self.assertEqual(changelist_url(widget),
                         admin_changelist_url + '?username__exact=user0')

        # Asserts first item is a Model
        widget.changelist_url = (None, {'username__exact': 'user0'})
        with self.assertRaises(AssertionError):
            self.assertEqual(changelist_url(widget), admin_changelist_url)

        # Asserts last items is either basestring or dict
        widget.changelist_url = (User, None)
        with self.assertRaises(AssertionError):
            self.assertEqual(changelist_url(widget), admin_changelist_url)

    def test_method_prop(self):
        class Test(object):
            foo = True

            def bar(self):
                pass
            bar.allow_tags = True

            def baz(self):
                pass
            baz.allow_tags = False

            def egg(self):
                pass

        test = Test()

        # Attribute is not callable
        self.assertIsNone(_method_prop(test, 'foo', 'allow_tags'))

        # Has the property
        self.assertEqual(_method_prop(test, 'bar', 'allow_tags'), True)

        # Has it but it's False
        self.assertFalse(_method_prop(test, 'baz', 'allow_tags'))

        # Doesn't have
        self.assertIsNone(_method_prop(test, 'egg', 'allow_tags'))

        # Doesn't exist
        self.assertIsNone(_method_prop(test, 'doesnt_exist', 'allow_tags'))


class AttrTagsTest(TestCase):
    def setUp(self):
        class TestUserWidget0(widgets.ItemList):
            model = TestUser0
            list_display = ('foo', 'egg')

            # Should override models method
            def foo(self, obj):
                return 'new foo value'
            foo.short_description = 'new foo label'

            # Doesn't have description
            def bar(self, obj):
                return 'new bar value'

            def allows_tags(self, obj):
                return '<br>'
            allows_tags.allow_tags = True

            def no_tags(self, obj):
                return '<br>'

        class TestUserWidget1(TestUserWidget0):
            list_display = None

        class TestUserWidget2(TestUserWidget0):
            list_display = ((app_settings.SHARP, ) +
                            TestUserWidget0.list_display)

        class TestUserWidget3(TestUserWidget2):
            model = TestUser1

        self.user0 = TestUser0(username='user0')
        self.widget0 = TestUserWidget0(request=None)
        self.widget1 = TestUserWidget1(request=None)
        self.widget2 = TestUserWidget2(request=None)
        self.widget3 = TestUserWidget3(request=None)

        self.mapping = {'baz': 'mapping baz'}
        self.sequence = ['foo value', 'egg value']
        self.namedtuple = collections.namedtuple('User', ['egg'])('egg value')

    def test_attrlabel(self):
        # Widget overrides
        self.assertEqual(attrlabel(self.widget0, 'foo'), 'new foo label')

        # Widget's has no description, takes model's one
        self.assertEqual(attrlabel(self.widget0, 'bar'), 'original bar label')

        # Empty description
        self.assertEqual(attrlabel(self.widget0, 'baz'), '')

        # Field's verbose name
        self.assertEqual(attrlabel(self.widget0, 'test_field'), 'My title')

        # No description found
        self.assertEqual(attrlabel(self.widget0, 'egg'), 'egg')

        # No attribute found
        self.assertEqual(attrlabel(self.widget0, 'unknown'), 'unknown')

        # Pk field
        self.assertEqual(attrlabel(self.widget0, 'id'), 'ID')
        self.assertEqual(attrlabel(self.widget0, 'pk'), 'ID')

        # Id is not defined
        self.assertEqual(attrlabel(self.widget3, 'id'), 'id')
        self.assertEqual(attrlabel(self.widget3, 'pk'), 'primary')

    def test_attrvalue(self):
        # New method
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'foo'), 'new foo value')

        # Old method
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'egg'), 'original egg value')

        # Allow tags test
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'allows_tags'), '<br>')
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'no_tags'), '&lt;br&gt;')

        # Attribute test
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'username'), 'user0')

        # 1) if method wasn't found in widget,
        #    doesn't pass instance to it's method
        # 2) returns empty value because gots None
        self.assertEqual(attrvalue(self.widget0, self.user0, 'baz'), '')

        # No attribute found -- empty value
        self.assertEqual(
            attrvalue(self.widget0, self.user0, 'unknown'), '')

        # Mapping test
        self.assertEqual(
            attrvalue(self.widget0, self.mapping, 'baz'), 'mapping baz')

        # Key not found, not KeyError
        self.assertEqual(
            attrvalue(self.widget0, self.mapping, 'unknown'), '')

        # Requires list_display to map it to values
        self.assertEqual(
            attrvalue(self.widget0, self.sequence, 'egg'), 'egg value')
        self.assertEqual(attrvalue(self.widget1, self.sequence, 'egg'), '')

        # Namedtuple doesn't require it
        # with list_display
        self.assertEqual(
            attrvalue(self.widget0, self.namedtuple, 'egg'), 'egg value')
        # without list_display
        self.assertEqual(
            attrvalue(self.widget1, self.namedtuple, 'egg'), 'egg value')

        # Sharp test
        self.assertEqual(
            attrvalue(self.widget2, self.sequence, 'egg'), 'egg value')

        # IndexError test
        self.assertEqual(
            attrvalue(self.widget2, self.sequence[:-1], 'egg'), '')


class ChangeurlTest(TestCase):
    def setUp(self):
        for i in range(10):
            username = 'user{}'.format(i)
            User.objects.create_user(username, username + '@example.com',
                                     username + 'password')

        self.obj = User.objects.first()
        self.obj_url = '/admin/auth/user/{}/'.format(self.obj.pk)
        if VERSION > (1, 9):
            self.obj_url += 'change/'

        # Model queryset
        class ModelQuerySet(widgets.ItemList):
            queryset = User.objects.all()

        # Deferred queryset
        class DeferredQuerySet(widgets.ItemList):
            queryset = User.objects.defer('email')

        # Dict
        class ValuesDict(widgets.ItemList):
            queryset = User.objects.values('pk', 'email')

        # List
        class ValuesList(widgets.ItemList):
            queryset = User.objects.values_list('pk', 'email')

        # List
        class ValuesListNoPk(widgets.ItemList):
            queryset = User.objects.values_list('email')

        # Namedtuple
        class NamedtupleList(ValuesList):
            klass = collections.namedtuple('User', 'pk email')

            def values(self):
                vals = super(NamedtupleList, self).values
                return [self.klass._make(x) for x in vals]

        self.widgets = [
            ModelQuerySet,
            DeferredQuerySet,
            ValuesDict,
            ValuesList,
            NamedtupleList,
        ]

        for widget in self.widgets:
            setattr(self, widget.__name__, widget)

    def equal(self, klass, value):
        widget = klass(request=None)
        self.assertEqual(change_url(widget, widget.values[0]), value)

    def test_non_registered(self):
        # It's not registered so no reverse is possible
        class NonRegisteredModel(widgets.ItemList):
            queryset = ContentType.objects.all()

        self.equal(NonRegisteredModel, None)

    def test_no_model(self):
        # Model queryset + Deferred
        self.equal(self.ModelQuerySet, self.obj_url)
        self.equal(self.DeferredQuerySet, self.obj_url)

        # widget.model is not defined, so it can't build
        # change_url from Dict, List, Namedtuple
        self.equal(self.ValuesDict, None)
        self.equal(self.ValuesList, None)
        self.equal(self.NamedtupleList, None)

    def test_with_model(self):
        for widget in self.widgets:
            class Widget(widget):
                model = User

            if widget is self.ValuesList:
                # No widget.values_list_defined
                self.equal(Widget, None)
            else:
                self.equal(Widget, self.obj_url)

    def test_with_model_and_list_display(self):
        for widget in self.widgets:
            class Widget(widget):
                model = User
                list_display = (app_settings.SHARP, 'pk', 'email')

            # Need pk to build url for ValuesList
            self.equal(Widget, self.obj_url)

            class IdWidget(Widget):
                list_display = ('id', 'email')

            # Alias test pk == id and also no sharp sign in list_display
            self.equal(IdWidget, self.obj_url)

    def test_no_pk(self):
        class NoPkList(self.NamedtupleList):
            klass = collections.namedtuple('User', 'email')
            model = User
            queryset = model.objects.values_list('email')

        self.equal(NoPkList, None)


class ExternalLinkTest(TestCase):
    def test_no_label(self):
        self.assertEqual(
            external_link('http://example.com'),
            '<a href="http://example.com" target="_blank" '
            'rel="noreferrer" rel="noopener">http://example.com</a>',
        )

    def test_with_label(self):
        self.assertEqual(
            external_link('http://example.com', 'my-example-link'),
            '<a href="http://example.com" target="_blank" '
            'rel="noreferrer" rel="noopener">my-example-link</a>',
        )



from controlcenter.base import BaseModel

from . import TestCase


class BaseModelTest(TestCase):
    def setUp(self):
        class NoTitleModel(BaseModel):
            pass

        class TitleModel(BaseModel):
            title = 'My Model'

        self.no_title = NoTitleModel()
        self.title = TitleModel()

    def test_no_title(self):
        self.assertEqual(self.no_title.__name__, 'NoTitleModel')
        self.assertEqual(self.no_title.slug, 'notitlemodel')
        self.assertEqual(self.no_title.title, 'No title model')

    def test_with_title(self):
        self.assertEqual(self.title.__name__, 'TitleModel')
        self.assertEqual(self.title.slug, 'titlemodel')
        self.assertEqual(self.title.title, 'My Model')

    def test_str_method(self):
        self.assertEqual(str(self.no_title), 'NoTitleModel')
        self.assertEqual(str(self.title), 'TitleModel')



from collections.abc import Sequence

from django.forms.widgets import MediaDefiningClass
from django.urls import reverse

from . import app_settings
from .base import BaseModel
from .widgets import Group

__all__ = ['Dashboard']


class Dashboard(BaseModel, metaclass=MediaDefiningClass):
    pk = None
    widgets = ()

    class Media:
        css = {
            'all': [
                'controlcenter/css/chartist.css',
                # This must follow chartist.css to override correctly:
                ('controlcenter/css/chartist-{}-colors.css'
                 .format(app_settings.CHARTIST_COLORS)),
                'controlcenter/css/all.css',
            ]
        }
        js = (
            'controlcenter/js/masonry.pkgd.min.js',
            'controlcenter/js/chartist/chartist.min.js',
            'controlcenter/js/chartist/chartist-plugin-pointlabels.min.js',
            'controlcenter/js/sortable.min.js',
            'controlcenter/js/scripts.js',
        )

    def __init__(self, pk):
        super(Dashboard, self).__init__()
        self.pk = self.id = pk

    def get_absolute_url(self):
        return reverse('controlcenter:dashboard', kwargs={'pk': self.pk})

    def get_widgets(self, request, **options):
        # TODO: permission check
        for item in self.widgets:
            if isinstance(item, Sequence):
                group = Group() + item
            else:
                group = Group([item])

            widgets = (x(request, **options) for x in group)
            new_group = Group(widgets, group.attrs, group.width, group.height)
            yield new_group



from controlcenter import Dashboard, widgets


class EmptyDashboard(Dashboard):
    pass


class MyWidget0(widgets.Widget):
    template_name = 'chart.html'


class MyWidget1(widgets.Widget):
    template_name = 'chart.html'


class NonEmptyDashboard(Dashboard):
    widgets = [
        MyWidget0,
        widgets.Group([MyWidget1])
    ]



from django.test.utils import override_settings

from controlcenter import Dashboard, app_settings

from . import TestCase


class AppConfTest(TestCase):
    def test_setup(self):
        self.assertEqual(app_settings.__name__, 'controlcenter.app_settings')

    def test_defaults(self):
        # Settings.py should not contain any controlcenter setting
        self.assertEqual(app_settings.DASHBOARDS, [])
        self.assertEqual(app_settings.CHARTIST_COLORS, 'default')
        self.assertEqual(app_settings.SHARP, '#')

    @override_settings(
        CONTROLCENTER_CHARTIST_COLORS='google',
        CONTROLCENTER_DASHBOARDS=[Dashboard])
    def test_changes(self):
        self.assertEqual(app_settings.DASHBOARDS, [Dashboard])
        self.assertEqual(app_settings.CHARTIST_COLORS, 'google')

    def test_unknown(self):
        with self.assertRaises(AttributeError):
            app_settings.DEBUG

    def test_monkeypatching(self):
        from django.conf import settings
        settings.CONTROLCENTER_CHARTIST_COLORS = 'monkey'
        self.assertEqual(app_settings.CHARTIST_COLORS, 'monkey')



from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.test.utils import override_settings

from controlcenter import widgets

from . import TestCase

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


@override_settings(CONTROLCENTER_DASHBOARDS=())
class A_DashboardTest(TestCase):
    # Must be first
    # 'cause @override_settings doesn't override settings well :(

    def setUp(self):
        self.superuser = User.objects.create_superuser(
            'superuser', 'superuser@example.com', 'superpassword')
        self.user = User.objects.create_user(
            'user', 'user@example.com', 'password')

    def test_no_dashboard_found(self):
        self.client.login(username='superuser', password='superpassword')
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(reverse('controlcenter:dashboard',
                                    kwargs={'pk': 0}))


class B_DashboardTest(TestCase):
    def setUp(self):
        self.superuser = User.objects.create_superuser(
            'superuser', 'superuser@example.com', 'superpassword')
        self.user = User.objects.create_user(
            'user', 'user@example.com', 'password')

    @override_settings(
        ROOT_URLCONF='urls',
        CONTROLCENTER_DASHBOARDS=[('empty', 'dashboards.EmptyDashboard')])
    def test_empty_dashboard(self):
        self.client.login(username='superuser', password='superpassword')

        # I wish I could cache urls, but reverse_lazy fails with py34 & 1.8
        # https://code.djangoproject.com/ticket/25424
        url_0 = reverse('controlcenter:dashboard', kwargs={'pk': 'empty'})
        response_0 = self.client.get(url_0)

        # Status code test
        self.assertEqual(response_0.status_code, 200)

        # Widgets test
        # It's empty, remember?
        # We don't check context, because `get_widgets` returns iterator
        # and it's empty now
        dashboard = response_0.context['dashboard']
        with self.assertRaises(StopIteration):
            next(dashboard.get_widgets(request=None))

        # Absolute url test
        self.assertEqual(dashboard.get_absolute_url(), url_0)

        # Non-exists dashboard test
        url_1 = reverse('controlcenter:dashboard', kwargs={'pk': 1})
        response_1 = self.client.get(url_1)
        self.assertEqual(response_1.status_code, 404)

    @override_settings(
        CONTROLCENTER_DASHBOARDS=('dashboards.EmptyDashboard',
                                  'dashboards.NonEmptyDashboard'))
    def test_regular_dashboard(self):
        url = reverse('controlcenter:dashboard', kwargs={'pk': 1})
        # Non-staff fails
        self.client.login(username='user', password='password')
        response = self.client.get(url)
        self.assertRedirects(response,
                             '{}?next={}'.format(reverse('admin:login'),
                                                 url))
        # Staff proceed
        self.client.login(username='superuser', password='superpassword')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        dashboard = response.context['dashboard']
        self.assertEqual(dashboard.get_absolute_url(), url)

        # It's not empty
        self.assertIsInstance(next(dashboard.get_widgets(request=None)),
                              widgets.Group)

    def test_multiple_dashboards(self):
        self.client.login(username='superuser', password='superpassword')
        dashboards = ['dashboards.NonEmptyDashboard' for i in range(30)]
        with self.settings(CONTROLCENTER_DASHBOARDS=dashboards):
            for i, path in enumerate(dashboards):
                url = reverse('controlcenter:dashboard', kwargs={'pk': i})
                response = self.client.get(url)
                self.assertEqual(response.status_code, 200)

    def test_navigation_links(self):
        self.client.login(username='superuser', password='superpassword')
        dashboards = (
            ('foo', 'dashboards.NonEmptyDashboard'),
            ('bar', 'dashboards.NonEmptyDashboard'),
            ('baz', 'dashboards.NonEmptyDashboard'),
        )
        with self.settings(CONTROLCENTER_DASHBOARDS=dashboards):
            response = self.client.get('/admin/dashboard/foo/')
            self.assertEqual(response.status_code, 200)
            expected = (
                '<div class="controlcenter__nav__item '
                'controlcenter__nav__item--active">Non empty dashboard</div>',
                '<a class="controlcenter__nav__item" '
                'href="/admin/dashboard/bar/">Non empty dashboard</a>',
                '<a class="controlcenter__nav__item" '
                'href="/admin/dashboard/baz/">Non empty dashboard</a>',
            )

            content = response.content.decode()
            for chunk in expected:
                # fixme: use subtest someday
                self.assertInHTML(chunk, content)

    def test_redirect_to_first_dashboard(self):
        self.client.login(username='superuser', password='superpassword')
        dashboards = (
            ('foo', 'dashboards.EmptyDashboard'),
            ('bar', 'dashboards.EmptyDashboard'),
            ('baz', 'dashboards.EmptyDashboard'),
        )
        with self.settings(CONTROLCENTER_DASHBOARDS=dashboards):
            url = reverse('controlcenter:index')
            self.assertEqual(url, '/admin/dashboard/')
            expected_url = reverse('controlcenter:dashboard', kwargs={'pk':'foo'})
            self.assertEqual(expected_url, '/admin/dashboard/foo/')
            response = self.client.get(url)
            self.assertRedirects(response, expected_url)



from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured

from controlcenter.widgets.charts import (
    BAR,
    LINE,
    PIE,
    BarChart,
    Chart,
    Chartist,
    LineChart,
    PieChart,
    SingleBarChart,
    SingleLineChart,
    SinglePieChart,
)

from . import TestCase


class BaseTest(TestCase):
    def test_chartist(self):
        class Chartist1(Chartist):
            foo = 'foo0'
            bar = 'bar0'
            options = {
                'foo': 'foo0',
                'bar': 'bar0'
            }

        class Chartist2(Chartist):
            foo = 'foo1'
            options = {
                'foo': 'foo1',
                'baz': 'baz1'
            }

            def egg(self):
                return 'egg'

        chartist = Chartist()
        chartist.update(Chartist1)
        chartist.update(Chartist2)

        # Works like python inheritance
        self.assertEqual(chartist.foo, 'foo1')
        self.assertEqual(chartist.bar, 'bar0')

        # Except methods
        with self.assertRaises(AttributeError):
            chartist.egg()

        # We have tested deepmerge in utils
        self.assertItemsEqual(chartist.options, ['bar', 'baz', 'foo'])

    def test_chart(self):
        chart = Chart(request=None)
        self.assertEqual(chart.legend, [])
        self.assertEqual(chart.labels, [])
        self.assertEqual(chart.series, [])

        # Queryset and model are not defined
        with self.assertRaises(ImproperlyConfigured):
            self.assertEqual(chart.values, [])

    def test_chartmeta(self):
        class Chart0(Chart):
            class Chartist:
                foo = 'foo0'
                bar = 'bar0'

        class Chart1(Chart):
            class Chartist:
                foo = 'foo1'
                baz = 'baz1'

        class Chart2(Chart0, Chart1):
            pass

        chart = Chart2(request=None)

        # Class removed
        self.assertFalse(hasattr(chart, 'Chartist'))
        # Attribute created
        self.assertTrue(hasattr(chart, 'chartist'))

        # Chartist inheritance
        self.assertEqual(chart.chartist.foo, 'foo1')
        self.assertEqual(chart.chartist.bar, 'bar0')
        self.assertEqual(chart.chartist.baz, 'baz1')

    def test_linechart(self):
        chart = LineChart(request=None)
        self.assertEqual(chart.chartist.klass, LINE)

    def test_piechart(self):
        chart = PieChart(request=None)
        self.assertEqual(chart.chartist.klass, PIE)

    def test_barchart(self):
        chart = BarChart(request=None)
        self.assertEqual(chart.chartist.klass, BAR)


class WidgetTest(TestCase):
    def setUp(self):
        for i in range(10):
            username = 'user{}'.format(i)
            User.objects.create_user(username, username + '@example.com',
                                     username + 'password')

    def test_singlepiechart(self):
        class Chart0(SinglePieChart):
            model = User

        chart0 = Chart0(request=None)

        # values_list is not defined
        with self.assertRaises(AssertionError):
            chart0.values

        class Chart1(Chart0):
            values_list = ['pk', 'email']
            limit_to = None  # there is limit by default

        chart1 = Chart1(request=None)
        self.assertItemsEqual(chart1.labels,
                              User.objects.values_list('pk', flat=True))

        self.assertItemsEqual(chart1.series,
                              User.objects.values_list('email', flat=True))

        # Override default limit_to
        max_items = 3

        class Chart2(Chart1):
            limit_to = max_items

        chart2 = Chart2(request=None)
        self.assertItemsEqual(
            chart2.labels,
            User.objects.values_list('pk', flat=True)[:max_items])

    def test_singlebarchart(self):
        chart0 = SingleBarChart(request=None)
        self.assertTrue(chart0.chartist.options['distributeSeries'])

    def test_singlelinechart(self):
        class Chart0(SingleLineChart):
            model = User
            values_list = ['pk', 'email']
            limit_to = None

        chart0 = Chart0(request=None)
        self.assertItemsEqual(
            chart0.series,
            [list(User.objects.values_list('email', flat=True))])



from setuptools import find_packages, setup

with open('README.rst', 'r') as f:
    long_description = f.read()

setup(
    name='django-controlcenter',
    version='0.3.2',
    description='Set of widgets to build dashboards for your Django-project.',
    long_description=long_description,
    url='https://github.com/byashimov/django-controlcenter',
    author='Murad Byashimov',
    author_email='byashimov@gmail.com',
    packages=find_packages(
        exclude=['controlcenter.stylus', 'controlcenter.images']),
    include_package_data=True,
    license='BSD',
    install_requires=['django-pkgconf~=0.4.0'],
    keywords='django admin dashboard',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: Django',
        'Framework :: Django :: 1',
        'Framework :: Django :: 2',
        'Framework :: Django :: 3',
        'Framework :: Django :: 4',
    ],
)



from django.contrib import admin

try:
    from django.urls import path
except ImportError:
    from django.conf.urls import url as path

from controlcenter.views import controlcenter


urlpatterns = [
    path('admin/dashboard/', controlcenter.urls),
    path('admin/', admin.site.urls),
]



from ..utils import deepmerge
from .core import Widget, WidgetMeta

__all__ = ['LineChart', 'TimeSeriesChart', 'BarChart', 'PieChart',
           'SingleLineChart', 'SingleBarChart', 'SinglePieChart',
           'LINE', 'BAR', 'PIE']

# Chart types
PIE, BAR, LINE = 'Pie', 'Bar', 'Line'


class Chartist(object):
    def __init__(self):
        self.options = {}

    def update(self, obj):
        for key in dir(obj):
            value = getattr(obj, key)
            if key == 'options':
                self.options = deepmerge(self.options, value)
            elif not key.startswith('__') and not callable(value):
                setattr(self, key, value)


class ChartMeta(WidgetMeta):
    CACHED_ATTRS = WidgetMeta.CACHED_ATTRS + (
        'labels',  # chart x-axis labels
        'series',  # chart y-axis values
        'legend',  # chart legend
    )

    def __new__(mcs, name, bases, attrs):
        # Saves defined configuration
        chartist = attrs.pop('Chartist', None)

        new_class = super(ChartMeta, mcs).__new__(mcs, name, bases, attrs)
        new_class.chartist = Chartist()

        # Copies parents options
        for base in bases:
            if hasattr(base, 'chartist'):
                new_class.chartist.update(base.chartist)

        # Overrides inherited stuff
        if chartist:
            new_class.chartist.update(chartist)
        return new_class


class Chart(Widget, metaclass=ChartMeta):
    template_name = 'chart.html'

    class Chartist:
        klass = LINE
        scale = 'octave'

    def legend(self):
        # Legend for chart
        return []

    def labels(self):
        # List of x-axis labels
        # Do not return generator!
        return []

    def series(self):
        # List of y-axis values
        # Do not return generator!
        return []


class LineChart(Chart):
    class Chartist:
        point_labels = True
        options = {
            # In common cases you need something last ordered by descending,
            # setting `reverseData` all the time is just annoying
            'reverseData': True,
            'axisY': {
                'onlyInteger': True,
            },
            'fullWidth': True,
        }


class TimeSeriesChart(Chart):
    class Chartist:
        point_labels = True
        options = {
            'axisY': {
                'onlyInteger': True,  # Same default as LineChart.
            },
        }
        time_series = True
        timestamp_options = {}


class BarChart(Chart):
    class Chartist:
        klass = BAR


class PieChart(Chart):
    class Chartist:
        klass = PIE


class SinglePieChart(PieChart):
    values_list = None

    def labels(self):
        return [x for x, y in self.values]

    def series(self):
        return [y for x, y in self.values]

    def values(self):
        assert self.values_list, ('Please define {0}.values_list '
                                  'or override {0}.values'.format(self))
        queryset = self.get_queryset().values_list(*self.values_list)
        if self.limit_to:
            return queryset[:self.limit_to]
        return queryset


class SingleBarChart(SinglePieChart, BarChart):
    class Chartist:
        options = {
            'distributeSeries': True
        }


class SingleLineChart(SinglePieChart, LineChart):
    def series(self):
        vals = super(SingleLineChart, self).series
        return [vals]



"""
Generic widgets for `django-controlcenter` dashboards that don't require
model data.
"""

from abc import abstractmethod
from collections import namedtuple

from .. import core

__all__ = ['BaseSimpleWidget', 'ValueList', 'KeyValueList', 'DataItem']


DataItem = namedtuple('DataItem', 'label url help_text')
DataItem.__new__.__defaults__ = ('', '', '')


class BaseSimpleWidget(core.BaseWidget):
    width = core.MEDIUM
    template_name_prefix = 'controlcenter/widgets/contrib'

    @abstractmethod
    def get_data(self):
        """
        Should be implemented in a subclass.
        """


class ValueList(BaseSimpleWidget):
    template_name = 'value_list.html'
    value_column_label = None
    sortable = False

    def items(self):
        return self.get_data()


class KeyValueList(BaseSimpleWidget):
    template_name = 'key_value_list.html'
    key_column_label = None
    value_column_label = None
    sortable = False

    def items(self):
        return self.get_data().items()
